import iris
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import numpy as np
import iris.quickplot as qplt
import os
import xarray as xr
from matplotlib.gridspec import GridSpec
from esmvalcore.preprocessor import extract_region, extract_time, rolling_window_statistics, climate_statistics, seasonal_statistics, area_statistics, extract_month, extract_season
import pandas as pd
import logging
from datetime import datetime, timedelta
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.metrics import silhouette_score
from scipy.stats import kendalltau, pointbiserialr, ttest_ind, mannwhitneyu
import matplotlib.patches as patches
import json
import matplotlib.colors as mcolors
import matplotlib as mpl
import seaborn as sns
import scipy
from scipy import stats
from esmvaltool.diag_scripts.shared  import run_diagnostic, group_metadata
import yaml

current_dir = os.path.dirname(os.path.realpath(__file__))

logger = logging.getLogger(__name__)
FORMAT = "%(asctime)-15s %(message)s"
logging.basicConfig(format=FORMAT)
logging.getLogger().setLevel(logging.INFO)

def regrid_longitude_coord(cube):
    """Sorts the longitudes of the cubes from 0/360 degrees to -180/180"""
    coord = cube.coord("longitude")
    lon_extent = iris.coords.CoordExtent(coord, -180., 180., True, False)
    cube = cube.intersection(lon_extent)
    return cube

def extract_months_xarray(data, desired_months):
    """
    desired_months is a list of month numbers (int)
    it takes only the dates that are in the months of the
    desired months list
    """
    for i,m in enumerate(desired_months):
        if m>12:
            desired_months[i] -= 12
    data['time'] = pd.to_datetime(data['time'].values)
    # Create a boolean mask for the desired months
    mask = data['time'].dt.month.isin(desired_months)
    # Apply the mask to the DataArray to select only the desired months
    out_data = data.sel(time=mask)
    form_time = [str(i).split("T")[0] for i in out_data.time.values]
    return out_data.assign_coords(time=form_time)

def compute_particular_dates_clim(cube, dates):
    dates = dates.drop_vars([coord for coord in dates.coords if "time" not in coord])
    yearly_clim = seasonal_statistics(cube.to_iris())
    season_days={"DJF": 90, "MAM": 92, "JJA": 92, "SON": 91}
    repeated_data = np.repeat(yearly_clim.data, season_days[yearly_clim.coord("clim_season").points[0]], axis=0)
    if cube.name == "tas":
        cube -= repeated_data
    cube_analogs = cube.sel(time=dates.time)
    clim = regrid_longitude_coord((cube_analogs.to_iris()).collapsed("time", iris.analysis.MEAN))
    return clim

def dates_file_writer(dataset, file, cfg, var, dates_dataset, identifier, period, season, bounds, minimum_intrusion_points_pcnt, sahara_clim_roll, sahara_region, ta_levels, gh_levels):
    ilon,flon,ilat,flat = bounds
    cube = iris.load_cube(file) 
    # save subsets of theta and thick
    xr_p = xr.DataArray.from_iris(regrid_longitude_coord(extract_region(cube,ilon,flon,ilat,flat)))
    xr_p = xr_p.assign_coords(time=[str(i).split(" ")[0] for i in xr_p.time.values]).sel(time=dates_dataset)
    if not os.path.isdir(os.path.join(cfg["work_dir"],f"{var}_intrusion_days")):
        os.makedirs(os.path.join(cfg["work_dir"],f"{var}_intrusion_days"))
    xr_p.to_netcdf(os.path.join(cfg["work_dir"],f"{var}_intrusion_days",f"{dataset}_{var}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"))
    
def compute_continuous_intrusion_dates(dataset, cfg, vars_dict, period, season, bounds, identifier, minimum_intrusion_points_pcnt, sahara_clim_roll, sahara_region, ta_levels, gh_levels):
    theta = regrid_longitude_coord(extract_region(iris.load_cube(os.path.join(cfg["work_dir"],vars_dict["ta"][0]["filename"].split("/")[-1].replace("ta","avgtap"))), -30, 50, 25, 50))
    thick = regrid_longitude_coord(extract_region(iris.load_cube(os.path.join(cfg["work_dir"],vars_dict["zg"][0]["filename"].split("/")[-1].replace("zg","gh"))), -30, 50, 25, 50))
    logger.info("compute intrusion datess")
    theta_threshold = iris.load_cube(os.path.join(cfg["work_dir"],f'threshold_{vars_dict["ta"][0]["filename"].split("/")[-1].replace("ta","avgtap")}'))
    thick_threshold = iris.load_cube(os.path.join(cfg["work_dir"],f'threshold_{vars_dict["zg"][0]["filename"].split("/")[-1].replace("zg","gh")}'))
  
    counts = theta.copy()

    no_saharian_theta = theta.data < theta_threshold.data[:,np.newaxis,np.newaxis]
    no_saharian_thick = thick.data < thick_threshold.data[:,np.newaxis,np.newaxis] #5800
    no_saharian = np.ma.mask_or(no_saharian_theta, no_saharian_thick)
    saharian = no_saharian == False
    counts = theta.copy(saharian+0)

    counts = regrid_longitude_coord(counts)

    ilon,flon,ilat,flat = bounds
    logger.info(f"computing intrusion days for {period.replace('-','').replace('.','-')} {season} {ilon}.{flon}.{ilat}.{flat}")
    counts_region = regrid_longitude_coord(extract_region(counts, ilon,flon,ilat,flat))

    points=counts_region.coord("longitude").shape[0]*counts_region.coord("latitude").shape[0]
    counts_areasum = counts_region[:,0,0].copy(np.sum(counts_region.data, axis=(1,2)))

    minimum_intrusion_points = minimum_intrusion_points_pcnt/100*points
    dates_intrusion = xr.DataArray.from_iris(counts_areasum).where(xr.DataArray.from_iris(counts_areasum)>=minimum_intrusion_points,drop=True)
    dates_intrusion_values = [str(i).split(" ")[0] for i in dates_intrusion.time.values]
    logger.info(len(dates_intrusion_values))

    # # Calculate the differences between consecutive dates
    # diffs = []
    # for date0,date1 in zip(dates_intrusion_values[:-1],dates_intrusion_values[1:]):
    #     diffs.append((datetime(year=int(date1[:4]), month=int(date1[5:7]), day=int(date1[8:]))-datetime(year=int(date0[:4]), month=int(date0[5:7]), day=int(date0[8:]))).days)
    # diffs = np.array(diffs)
    # # Find the indices where the differences are not equal to 1
    # start_indices = np.where(diffs != 1)[0]

    # # List of spell first day
    # t0=[dates_intrusion_values[start_index] for start_index in start_indices]

    # # precursors array
    # t_minus1 = [i for i in [(datetime(year=int(date[:4]), month=int(date[5:7]), day=int(date[8:]))-timedelta(days=1)).strftime('%Y-%m-%d') for date in t0] if "-05-" not in i]
    # t_minus2 = [i for i in [(datetime(year=int(date[:4]), month=int(date[5:7]), day=int(date[8:]))-timedelta(days=2)).strftime('%Y-%m-%d') for date in t0] if "-05-" not in i]
    # t_minus1n2 = sorted(t_minus1+t_minus2)
    # t_all = sorted(t_minus1n2+list(dates_intrusion_values))
    # dates_intrusion_values_nminus1 = sorted(t_minus1+list(dates_intrusion_values))

    def _dates_file_writer(dates_dataset, identifier):
        if not os.path.isdir(os.path.join(cfg["work_dir"],f"intrusion_dates")):
            os.makedirs(os.path.join(cfg["work_dir"],f"intrusion_dates"))
        with open(os.path.join(cfg["work_dir"],f"intrusion_dates/intrusions-clim-dates_{dataset}_{ilon}.{flon}.{ilat}.{flat}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.txt"), "w") as f:
            logger.info(os.path.join(cfg["work_dir"],f"intrusion_dates/intrusions-clim-dates_{dataset}_{ilon}.{flon}.{ilat}.{flat}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.txt"))
            for string in dates_dataset:
                f.write(string + "\n")

    _dates_file_writer(dates_intrusion_values,"")
    return dates_intrusion_values
    # if identifier == "-only1":
    #     _dates_file_writer(t_minus1,"-only1")
    #     return t_minus1
    # if identifier == "-only2":
    #     _dates_file_writer(t_minus2,"-only2")
    #     return t_minus2
    # if identifier == "-only1n2":
    #     _dates_file_writer(t_minus1n2,"-only1n-2")
    #     return t_minus1n2
    # if identifier == "-all":
    #     _dates_file_writer(t_all,"-all")
    #     return t_all
    # if identifier == "":
    #     _dates_file_writer(dates_intrusion_values,"")
    #     return dates_intrusion_values
    # if identifier == "-1":
    #     _dates_file_writer(dates_intrusion_values_nminus1,"-1")
    #     return dates_intrusion_values_nminus1



def pseudof_study(dataset, cfg, data, period, season, bounds, identifier, minimum_intrusion_points_pcnt, indicators):
    concatenable_data = [data[i].stack(space=['lat', 'lon']) for i in range(len(data))]
    ilon,flon,ilat,flat = bounds
    if len(concatenable_data)==0:
        return None
    M=xr.concat(concatenable_data, dim="space").drop_vars(["lat", "lon"]).assign_coords(space=np.arange(concatenable_data[0].shape[1]*len(concatenable_data))).dropna(dim="space")

    from_var = ""
    if from_var == "psl-":
        psl_file = "/home/jcos/python_scripts/subtropical_intrusions/files/psl_194001-202212_MatiDM_.5x.5.nc"
        psl = xr.open_dataset(psl_file).psl
        psl = psl.assign_coords(time=[str(i).split(" ")[0] for i in psl.time.values]).sel(time=M.time)
        M=psl.stack(space=['lat', 'lon']).drop_vars(["lat", "lon"]).assign_coords(space=np.arange(psl.shape[1]*psl.shape[2])) 
    upper_bound = M.shape[0]
    if upper_bound>100:
        upper_bound=100
    n_modes = upper_bound
    pca_data = PCA(n_components=n_modes)
    pca_data.fit(M.values.T)
    scores = pca_data.components_
    expvar_ratio_sklr = pca_data.explained_variance_ratio_
    min_modes = np.argwhere(np.array([np.sum(expvar_ratio_sklr[:i+1]) for i in range(expvar_ratio_sklr.shape[0])]) > 0.99)[0][0]
    logger.info(f"min_modes: {min_modes}")
    # Use only needed modes
    clustering_matrix = scores[:min_modes+1].T

    overall_mean = np.mean(clustering_matrix, axis=0)
    logger.info(M.shape[0])

    if M.shape[0] < 15:
        K = range(2,M.shape[0])
    else:
        K = range(2, 15)
    # Pseudo-F
    psf_data = []
    sil = []
    wsss = []
    for k in K:
        km = KMeans(n_clusters=k, n_init=10, max_iter = 1000, random_state=0)
        km.fit(clustering_matrix)
        kmean_data = km.fit_predict(clustering_matrix)
        a=silhouette_score(clustering_matrix, kmean_data)
        sil.append(a)
        centroids = km.cluster_centers_
        centroids_mean = np.mean(centroids, axis=0)
        gss = np.sum([np.sum((clustering_matrix[kmean_data == k] - overall_mean) ** 2) for k in np.unique(kmean_data)])
        wss = np.sum([np.sum((clustering_matrix[kmean_data == k] - np.mean(clustering_matrix[kmean_data == k], axis=0)) ** 2) for k in np.unique(kmean_data)])
        df_between = len(np.unique(kmean_data)) - 1
        df_within = len(clustering_matrix) - len(np.unique(kmean_data))
        pf = (gss / df_between) / (wss / df_within)
        psf_data.append(pf)
        wsss.append(km.inertia_)

  

    fig, ax1 = plt.subplots()

    # Plot the first dataset (y1) on the left y-axis
    ax1.plot(K, psf_data, 'x-', color='g')
    ax1.set_ylabel('Pseudo-F', color='g')
    ax1.tick_params(axis='y', labelcolor='g')
    ax1.set_xlabel('k')

    ax2 = ax1.twinx()
    ax2.plot(K, sil, 'x-', color='k')
    ax2.set_ylabel('Silhouette score', color='k')
    ax2.tick_params(axis='y', labelcolor='k')
    ax2.set_title(f'{season.upper()} Pseudo-F and Silhouette Score for Optimal k')
    plt.tight_layout()
    plt.savefig(os.path.join(cfg["plot_dir"],f"{dataset}_{season}_pseudof.png"))


def find_max_16_blob(grid):
    rows, cols = grid.shape
    max_sum = -np.inf
    max_position = (0, 0)
    
    # Iterate over the grid with a 8x8 window
    for i in range(rows - 15):
        for j in range(cols - 15):
            # Extract the 4x4 window
            window = grid[i:i+16, j:j+16]
            window_sum = np.sum(window)
            
            # Update max_sum and max_position if current window_sum is greater
            if window_sum > max_sum:
                max_sum = window_sum
                max_position = (i, j)
    
    return max_sum, max_position


def taylored_clustering(dataset, cfg, var_dict, k, data, cluster_vars, period, season, bounds, identifier, minimum_intrusion_points_pcnt, sahara_clim_roll, sahara_region, ta_levels, gh_levels):
    concatenable_data = [data[i].stack(space=['lat', 'lon']) for i in range(len(data))]
    ilon,flon,ilat,flat = bounds
    M=xr.concat(concatenable_data, dim="space").drop_vars(["lat", "lon"]).assign_coords(space=np.arange(concatenable_data[0].shape[1]*len(concatenable_data))).dropna(dim="space")
    # CLUSTERS k
    upper_bound = min(M.shape)
    if upper_bound>100:
        upper_bound=100
    n_modes = upper_bound
    pca_data = PCA(n_components=n_modes)
    pca_data.fit(M.values.T)
    scores = pca_data.components_
    expvar_ratio_sklr = pca_data.explained_variance_ratio_
    min_modes = np.argwhere(np.array([np.sum(expvar_ratio_sklr[:i+1]) for i in range(expvar_ratio_sklr.shape[0])]) > 0.99)[0][0]
    logger.info(f"min_modes: {min_modes}")
    # Use only needed modes
    clustering_matrix = scores[:min_modes+1].T
    km = KMeans(n_clusters=k, n_init=100, max_iter = 1000, random_state=0)
    kmean_data = km.fit_predict(clustering_matrix)
    
    days={k_id: concatenable_data[0][kmean_data == k_id].time.values for k_id in np.unique(kmean_data)}

    lon_pos = [np.linalg.norm(find_max_16_blob(data[0][np.where(kmean_data==i)[0]].mean(axis=0).values)[1]-np.array([ilat,ilon])) for i in range(k)]
    logger.info(f"theta max pos: {lon_pos}")

    cp_id = np.argsort(np.array(lon_pos))
    logger.info(cp_id)
    maper = {cp: i for i, cp in enumerate(cp_id)}
    kmean_data = np.vectorize(maper.get)(kmean_data)
    if season == "djf":
        maper={0:0,1:3,2:1,3:2,4:4}
        kmean_data = np.vectorize(maper.get)(kmean_data)
    elif season == "son":
        # pass
        maper={0:1,1:0,2:2,3:3}
        kmean_data = np.vectorize(maper.get)(kmean_data)
    logger.info(f"maper: {maper}")

    cluster_vector = xr.DataArray(kmean_data, coords={"time": concatenable_data[0].time.values}, dims=["time"])
    cluster_vector.to_netcdf(os.path.join(cfg["work_dir"],f"{dataset}_clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"))

    file_theta = os.path.join(cfg["work_dir"],var_dict["ta"][0]["filename"].split("/")[-1].replace("ta", "avgtap"))
    file_thick = os.path.join(cfg["work_dir"],var_dict["zg"][0]["filename"].split("/")[-1].replace("zg", "gh"))

    theta = iris.load_cube(file_theta)
    thick = iris.load_cube(file_thick)
    theta_threshold = iris.load_cube(os.path.join(cfg["work_dir"],f'threshold_{var_dict["ta"][0]["filename"].split("/")[-1].replace("ta","avgtap")}'))
    thick_threshold = iris.load_cube(os.path.join(cfg["work_dir"],f'threshold_{var_dict["zg"][0]["filename"].split("/")[-1].replace("zg","gh")}'))
  
    counts = theta.copy()

    no_saharian_theta = theta.data < theta_threshold.data[:,np.newaxis,np.newaxis]
    no_saharian_thick = thick.data < thick_threshold.data[:,np.newaxis,np.newaxis] #5800
    no_saharian = np.ma.mask_or(no_saharian_theta, no_saharian_thick)
    saharian = no_saharian == False
    counts = theta.copy(saharian+0)

    # if not os.path.isdir(os.path.join(cfg["work_dir"],f"area_of_impact")):
    #     os.makedirs(os.path.join(cfg["work_dir"],f"area_of_impact"))
    # if not os.path.isdir(os.path.join(cfg["plot_dir"],f"area_of_impact")):
    #     os.makedirs(os.path.join(cfg["plot_dir"],f"area_of_impact"))

    # counts = xr.DataArray.from_iris(regrid_longitude_coord(counts))
    # for n in sorted(list(set(cluster_vector.values))):
    #     cluster_times = cluster_vector.where(cluster_vector==n, drop=True).time
    #     area_of_impact = counts.assign_coords(time=[str(i).split(" ")[0] for i in counts.time.values]).sel(time=cluster_times.values).sum(dim="time")
    #     area_of_impact = area_of_impact.where(area_of_impact!=0).to_iris()
    #     iris.save(area_of_impact,os.path.join(cfg["work_dir"], "area_of_impact", f"area-of-impact-n{n+1}_{dataset}_clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"))
    #     plt.figure()
    #     plt.subplot(111,projection=ccrs.PlateCarree())
    #     qplt.pcolormesh(area_of_impact)
    #     plt.gca().coastlines()
    #     plt.savefig(os.path.join(cfg["plot_dir"],f"area-of-impact-n{n+1}_{dataset}_{season}_clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.png"))

def compute_continuous_thresholds(dataset, alias_dict, cfg, sahara_clim_roll, sahara_region, ta_levels, gh_levels):
    lon_min, lon_max, lat_min, lat_max = sahara_region
    logger.info(f"{lon_min},{lon_max},{lat_min},{lat_max}")
    vars_dict = group_metadata(alias_dict[dataset],"short_name")
    theta_sahara = area_statistics(extract_region(iris.load_cube(os.path.join(cfg["work_dir"],vars_dict["ta"][0]["filename"].split("/")[-1].replace("ta","avgtap"))),lon_min,lon_max,lat_min,lat_max), "mean")
    thick_sahara = area_statistics(extract_region(iris.load_cube(os.path.join(cfg["work_dir"],vars_dict["zg"][0]["filename"].split("/")[-1].replace("zg", "gh"))),lon_min,lon_max,lat_min,lat_max), "mean")
    dts=theta_sahara.coord("time").units.num2date(theta_sahara.coord("time").points)
    theta_thresholds_cube_ann = climate_statistics(rolling_window_statistics(theta_sahara, "time", "mean", sahara_clim_roll),"mean","day")
    thick_thresholds_cube_ann = climate_statistics(rolling_window_statistics(thick_sahara, "time", "mean", sahara_clim_roll),"mean","day")
    theta_thresholds_data = []
    thick_thresholds_data = []
    for year in np.arange(int(dts[0].year), int(dts[-1].year)+1):
        theta_data = list(np.around(theta_thresholds_cube_ann.data,2))
        thick_data = list(np.around(thick_thresholds_cube_ann.data,2))
        theta_thresholds_data += theta_data
        thick_thresholds_data += thick_data
    theta_thresholds_cube = theta_sahara.copy(np.array(theta_thresholds_data))
    thick_thresholds_cube = thick_sahara.copy(np.array(thick_thresholds_data))
    iris.save(theta_thresholds_cube, os.path.join(cfg["work_dir"],f'threshold_{vars_dict["ta"][0]["filename"].split("/")[-1].replace("ta","avgtap")}'))
    iris.save(thick_thresholds_cube, os.path.join(cfg["work_dir"],f'threshold_{vars_dict["zg"][0]["filename"].split("/")[-1].replace("zg", "gh")}'))
   

def plot_area_of_impact(ax, area_of_impact, lstyle):
    """
    Plot the area of influence of the IT. 
    The green contour defines the threshold where at least one 
    intrusion day is found.
    """
    area_of_impact = area_of_impact.where(area_of_impact.values<1000)
    z_min = 1
    z_max = np.floor(area_of_impact.max() * 2) / 2
    step = (z_max - z_min) / 3
    step = np.round(step * 2) / 2 
    if int(step) == 0:
        step=1
    levels = np.arange(int(z_min), int(z_max + step), int(step))
    area_of_impact.plot.contour(levels=levels[:1], colors="green", linewidths=1.7, linestyles=lstyle, transform=ccrs.PlateCarree(), alpha=0.7)
    return ax



def plot_heatmap_boxes(fig, ax, cax, events_da):
    # Extract colors from the colormap
    cmap = cax.get_cmap()  # Get the colormap used in the colorbar
    norm = cax.norm  # Normalization (maps data values to 0-1 range)

    # Get the unique data values that are actually plotted
    unique_values = np.unique(events_da)[~np.isnan(np.unique(events_da))]

    normalized_values = norm(unique_values)

    # Map the normalized values to colors using the colormap
    colors = [cmap(val) for val in normalized_values]
    N = len(colors)  # Number of boxes

    # Disable the axes
    ax.set_axis_off()

    # Size of the square patches
    box_size = 0.7

    # Loop to create the boxes and labels
    for i in range(N):
        # Create a square patch (rectangle with equal width and height)
        rect = patches.Rectangle((i * (box_size + 0.2), 0), box_size, box_size, 
                                edgecolor='black', facecolor=colors[i])
        ax.add_patch(rect)
        
        # Add centered text inside each box with bold font
        ax.text(i * (box_size + 0.2) + box_size / 2, box_size / 2, f'IT{i+1}', 
                ha='center', va='center', fontsize=10, fontweight='bold', color='black',bbox=dict(facecolor='white', edgecolor='black'))

    # Adjust the plot limits to fit all boxes and text
    ax.set_xlim(-0.2, N * (box_size + 0.2))
    ax.set_ylim(0, box_size)

# plot intrusion events in pcolormesh
# with the cluster color
def episodes_ploter(dataset, cfg, var_dict, k, period, season, bounds, identifier, minimum_intrusion_points_pcnt, sahara_clim_roll, sahara_region, ta_levels, gh_levels, indicators, seasons_dict, cluster_vars=[""]):
    ilon,flon,ilat,flat = bounds
    ds=iris.load_cube(os.path.join(cfg["work_dir"],var_dict["ta"][0]["filename"].split("/")[-1].replace("ta", "avgtap")))
    if season != "ann" and isinstance(season, str):
        fs=(8.5,8.5)
        logger.info(f"extracting season {season}")
        ds_s = xr.DataArray.from_iris(extract_season(ds, season))
        size="25%"
    elif isinstance(season, int) or isinstance(season, np.int64):
        fs=(8,5.5)
        logger.info(f"extracting month {season}")
        ds_s = xr.DataArray.from_iris(extract_month(ds, season))
        size="25%"
    else:
        fs=(9,33)
        logger.info("ANNUAL")
        ds_s = xr.DataArray.from_iris(ds)
        size="10%"
    all_time_vector = [str(i).split(" ")[0] for i in ds_s.time.values]
    years = sorted(list(set([i.split("-")[0] for i in all_time_vector])))
    days = sorted(list(set(["-".join(i.split("-")[1:]) for i in all_time_vector])))
    if season == "djf":
        days = days[60:]+days[0:60]
        years = [1958]+years

    clusters_list = np.ones((len(days), len(years)))*-1
    clusters = xr.open_dataarray(os.path.join(cfg["work_dir"],f"{dataset}_clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"))
    print(clusters)
    for j, d in enumerate(days):
        iterator = years
        if "dj" in season:
            iterator = years[1:]
        for i, y in enumerate(iterator):
            date = f"{y}-{d}"
            if date in clusters.time.values:
                if "-12-" in date and "dj" in season:
                    clusters_list[j][i+1] = int(clusters.sel(time=date).values) 
                else:
                    clusters_list[j][i] = int(clusters.sel(time=date).values) 
   
    days=[f'{i.split("-")[-1]}-{i.split("-")[0]}' for i in days]
    events_da = xr.DataArray(clusters_list, coords=[days, years], dims=["days", 'years'], name='clusters')+1
    months=list(set([i.split("-")[-1] for i in days]))

    # Yearly intrusion days
    yearly_intrusion_days = [] 
    for cluster_id in range(1,k+1):
        events_pre = events_da.copy()
        events_pre.values[events_da.values != cluster_id] = np.nan
        event_ones=events_pre-events_pre+1
        yearly_intrusion_days.append(np.sum(event_ones, axis=0))

    # Monthly intrusion days
    ev=events_da.copy()
    ev.values[events_da.values == 0] = np.nan
    ev_ones=ev-ev+1
    monthly_intrusion_days = {}
    for mth in months:
        monthly_intrusion_days[mth]=np.nansum(np.array([[i.values for i in ev_ones[:,k] if f"-{mth}" in str(i.days.values)] for k in range(ev_ones.shape[1])]), axis=1)


    events_da.values[events_da.values == 0]= np.nan
    event_ones=events_da-events_da+1
    yearly_intrusion_days.append(np.sum(event_ones, axis=0))

    yearly_intrusion_days_dataset = xr.Dataset({f"IT{n+1}": yearly_intrusion_days[n] for n in range(k+1)})
    yearly_intrusion_days = yearly_intrusion_days[-1]

    # # FUNCTION
    # # JJALEMAN
    # # Group 1: [1940, 1945, 1947, 1950, 1960, 1982, 1992, 1993, 1995, 2003, 2009, 2015, 2017, 2022]
    # # Group 2: [1941, 1944, 1948, 1954, 1960, 1965, 1966, 1970, 1972, 1973, 1977, 1978, 1979, 1980, 1991, 1996, 1998, 2002, 2005, 2012, 2014, 2021] 
    # g1=[1982, 1992, 1993, 1995, 2003, 2009, 2015, 2017, 2022]
    # g2=[1965, 1966, 1970, 1972, 1973, 1977, 1978, 1979, 1980, 1991, 1996, 1998, 2002, 2005, 2012, 2014, 2021] 
    # g1_bin = np.zeros(yearly_intrusion_days_dataset.years.shape)
    # g2_bin = np.zeros(yearly_intrusion_days_dataset.years.shape)
    # for i, g in enumerate(yearly_intrusion_days_dataset.years):
    #     if int(g) in g1: 
    #         g1_bin[i] = 1
    #     if int(g) in g2: 
    #         g2_bin[i] = 1

    # jjdict = {}
    # for it in yearly_intrusion_days_dataset:
    #     correlation_g2, p_value_g2 = pointbiserialr(g2_bin, yearly_intrusion_days_dataset[it].values)
    #     correlation_g1, p_value_g1 = pointbiserialr(g1_bin, yearly_intrusion_days_dataset[it].values)
    #     tau_g2, p_value_kendall_g2 = kendalltau(g2_bin, yearly_intrusion_days_dataset[it].values)
    #     tau_g1, p_value_kendall_g1 = kendalltau(g1_bin, yearly_intrusion_days_dataset[it].values)

    #     events_with_property = np.array(yearly_intrusion_days_dataset[it].values)[np.array(g1_bin) == 1]
    #     events_without_property = np.array(yearly_intrusion_days_dataset[it].values)[np.array(g1_bin) == 0]
    #     t_stat_g1, p_value_g1tt = ttest_ind(events_with_property, events_without_property, equal_var=False)
    #     u_stat_g1, p_value_g1u = mannwhitneyu(events_with_property, events_without_property)
    #     events_with_property = np.array(yearly_intrusion_days_dataset[it].values)[np.array(g2_bin) == 1]
    #     events_without_property = np.array(yearly_intrusion_days_dataset[it].values)[np.array(g2_bin) == 0]
    #     t_stat_g2, p_value_g2tt = ttest_ind(events_with_property, events_without_property, equal_var=False)
    #     u_stat_g2, p_value_g2u = mannwhitneyu(events_with_property, events_without_property)


    #     jjdict[f"{season}_{it}_g1_kendall"] = [tau_g1, p_value_kendall_g1]
    #     jjdict[f"{season}_{it}_g2_kendall"] = [tau_g2, p_value_kendall_g2]
    #     jjdict[f"{season}_{it}_g1_biserial"] = [correlation_g1, p_value_g1]
    #     jjdict[f"{season}_{it}_g2_biserial"] = [correlation_g2, p_value_g2]
    #     jjdict[f"{season}_{it}_g1_tt"] = [t_stat_g1, p_value_g1tt]
    #     jjdict[f"{season}_{it}_g2_tt"] = [t_stat_g2, p_value_g2tt]
    #     jjdict[f"{season}_{it}_g1_u"] = [t_stat_g1, p_value_g1u]
    #     jjdict[f"{season}_{it}_g2_u"] = [t_stat_g2, p_value_g2u]
    #     logger.info(f"{season} {it}:\nbiser g1 {correlation_g1}, {p_value_g1}\nbiser g2 {correlation_g2}, {p_value_g2}\ntau g1 {tau_g1}, {p_value_kendall_g1}\ntau g2 {tau_g2}, {p_value_kendall_g2}\ntt g1 {t_stat_g1}, {p_value_g1tt}\ntt g2 {t_stat_g2}, {p_value_g2tt}\nu g1 {u_stat_g1}, {p_value_g1u}\nu g2 {u_stat_g2}, {p_value_g2u}")
    # with open(os.path.join(cfg["work_dir"], f"files/postpro_files/gs_{season}.json"), "w") as f:
    #     json.dump(jjdict, f)
    # # END FUNC


    # Use a mask to identify where the values are 1
    ones_mask = np.isnan(event_ones) == False  # True where there are ones or NaNs
    # Find the indices where the array changes from NaN to 1 or 1 to NaN
    change_indices = np.diff(ones_mask.astype(int),axis=0)
    lengths = []
    for y in range(change_indices.shape[-1]):
        # Identify the start and end of each run of ones
        starts = np.where(change_indices[:,y] == 1)[0] + 1
        ends = np.where(change_indices[:,y] == -1)[0] + 1
        # If the array ends with a run of ones, include the last index
        if event_ones[-1,y] == 1:
            ends = np.append(ends, len(event_ones[:,y]))
        if event_ones[0,y] == 1:
            starts = np.insert(starts, 0,0)
            
        # Calculate the lengths of the runs of ones
        if len(ends) != len(starts):
            logger.info(starts)
            logger.info(ends)
            logger.info(f"{y}")
        lengths.append(np.mean(ends - starts))
    lengths=[0 if np.isnan(x) else x for x in lengths]

    palettes = [
        ["#FF0000", "#00FF00", "#0000FF", "#FFFF00", "#FF00FF"],  # High-contrast RGB
        ["#7F0000", "#7FFF00", "#007FFF", "#FF7F00", "#7F7FFF"],  # Variants of RGB
        ["#4B0082", "#8B0000", "#FF4500", "#32CD32", "#00CED1"],  # Contrasting diverse hues
        ["#FFD700", "#ADFF2F", "#FF69B4", "#4682B4", "#8A2BE2"],  # Vibrant distinct colors
        ["#FFD700", "#ADFF2F", "#FF69B4", "#4682B4", "#8A2BE2"],  # Vibrant distinct colors
        ["#FFD700", "#ADFF2F", "#FF69B4", "#4682B4", "#8A2BE2"],  # Vibrant distinct colors
        ["#FFD700", "#ADFF2F", "#FF69B4", "#4682B4", "#8A2BE2"],  # Vibrant distinct colors
    ]

    fig = plt.figure(figsize=(fs[0], fs[1]+fs[1]*0.05))
    gs = GridSpec(2, 1, height_ratios=[100, 5])
    ax = plt.subplot(gs[0])
    if season == "ann":
        vmin, vmax= [-1,1]
        cax = events_da.plot(edgecolors="whitesmoke", linewidth=0.05, vmin=vmin, vmax=vmax, cmap=plt.cm.RdBu_r, ax=ax,add_colorbar=False)
    else:
        addition = [i for i, s in enumerate(seasons_dict) if s == season][0]
        vmin, vmax= np.array([-(k+.5),k])
        rainbow_cmap = mcolors.ListedColormap(["#FFFFFF"]+palettes[addition][:k], name="rainbow_discrete")
        bounds = np.arange(-0.5,k+0.6)
        norm = mpl.colors.BoundaryNorm(bounds, rainbow_cmap.N)
        # vmin, vmax= [0,0.5]
        cax = events_da.plot(edgecolors="whitesmoke", linewidth=0.05, cmap=rainbow_cmap, norm=norm, ax=ax,add_colorbar=False)
    months = set([i.split("-")[0] for i in days])
    start_year="1959"
    if season == "djf":
        start_year="1958"
    for month in months:
        plt.hlines(f"01-{month}", start_year, "2022", colors="black")
    if season == "djf":
        plt.xticks(ticks=ax.get_xticks()[::2], labels=[f"{i}/{str(int(i)+1)[-2:]}" for i in years][::2], rotation=90)
    else:
        plt.xticks(ticks=years[::2], labels=years[::2], rotation=90)
    plt.yticks(ticks=days[::3], labels=days[::3])

    # if "mjj" in season:
    #     plt.ylim(plt.ylim()[0], plt.ylim()[1])
    #     for y1 in g1:
    #         plt.vlines(str(y1), plt.ylim()[0], plt.ylim()[1], color="blue", alpha=0.3)
    #     for y2 in g2:
    #         plt.vlines(str(y2), plt.ylim()[0], plt.ylim()[1], color="red", alpha=0.3)

    ax2 = plt.subplot(gs[1])
    plot_heatmap_boxes(fig, ax2, cax, events_da)
    plt.tight_layout()
    plt.savefig(os.path.join(cfg["plot_dir"], f"{dataset}_ind{indicators}_FigureS7-5{seasons_dict[season]}_cb.png"))

    # divider = make_axes_locatable(ax)
    # ax3 = divider.append_axes("bottom", size=size, pad=0.1, axes_class=plt.Axes)
    if season == "djf":
        years = years[1:-1]
        yearly_intrusion_days = yearly_intrusion_days[1:-1]
        lengths = lengths[1:-1]
    slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(np.array(years, dtype=float), np.array(yearly_intrusion_days.values, dtype=float))
    res = stats.kendalltau(np.array(years, dtype=float), np.array(yearly_intrusion_days.values, dtype=float))
    pval = res.pvalue
    fig = plt.figure(figsize=(8.5*0.85+8.5*0.85/5,2.5*0.9), dpi=200)  # Adjust the figsize as needed
    gs = GridSpec(1, 6, figure=fig)  #
    ax3 = fig.add_subplot(gs[0, :5])
    ax3.bar(years, yearly_intrusion_days.values, label="number of days")
    ax3.bar(years, lengths, label="event mean persistence")
    ax3.plot(years, intercept + slope * np.array(years, dtype=float), label=f"slope: {slope:.2f}, r: {r_value:.2f}, Mann-Kendall p-value: {pval:.4f}")
    ax3.set_xlim(years[0], years[-1])
    ax3.set_xticks(years[::2])
    if season == "djf":
        ax3.set_xticklabels([f"{i}/{str(int(i)+1)[-2:]}"for i in years][::2])
    ax3.set_ylim(-0.5,32)
    ax3.hlines(0,ax3.get_xlim()[0], ax3.get_xlim()[1], colors="black", alpha=0.5)
    plt.xticks(rotation=90)
    plt.legend()
    ax3.set_xlabel("years")
    ax3.set_ylabel("days")
    if season != "ann" and "mjj" not in season:
        ax = fig.add_subplot(gs[0, 5], sharey=ax3) 
        ax.tick_params(left=False, labelleft=False)
        months = ["Jan.", "Feb.", "Mar.", "Apr.", "May.", "Jun.", "Jul.", "Aug.", "Sep.", "Oct.", "Nov.", "Dec."]*2
        allmonths = "jfmamjjasond" * 2
        sstart = allmonths.index(season)
        desired_months = [months[i-1] for i in range(sstart+1,sstart+1+len(season))]
        keys = sorted(list(monthly_intrusion_days.keys()))
        if season == "djf":
            keys=keys[-1:]+keys[:-1]
            bxp = [monthly_intrusion_days[i][1:-1] for i in keys]
        else:
            bxp = [monthly_intrusion_days[i] for i in keys]
        bxp.append(yearly_intrusion_days.values)
        medianprops = dict(color='#1f77b4', linewidth=1.2)  # Change median line color
        flierprops = dict(marker='o', color='black', markersize=3)  # Change outlier size & color
        # plt.boxplot(bxp, positions=[0,.2,.4,.6], labels=[f"{j}" for j in desired_months]+[f"{season.upper()}"], medianprops=medianprops, flierprops=flierprops)
        positions=np.linspace(0,1.5,4)
        vp = ax.violinplot(bxp, positions=positions) #, labels=[f"{j}" for j in desired_months]+[f"{season.upper()}"]) #, medianprops=medianprops, flierprops=flierprops)
        for partname in ('cbars', 'cmins', 'cmaxes'):  # Customizing boxplot lines
            vp[partname].set_linewidth(1)
        ax.set_xticks(positions, labels=[f"{j}" for j in desired_months]+[f"{season.upper()}"])
        ax.set_xlim(positions[0]-(positions[1]-positions[0])/2,positions[-1]+(positions[1]-positions[0])/2)
        ax.tick_params(axis="x", rotation=90)
    ax3.hlines(0,ax.get_xlim()[0], ax.get_xlim()[1], colors="black", alpha=0.5)
    plt.tight_layout()
    plt.savefig(os.path.join(cfg["plot_dir"], f"{dataset}_ind{indicators}_Figure6{seasons_dict[season]}.png"))

    df = yearly_intrusion_days_dataset.to_dataframe().reset_index() 
    df.to_csv(os.path.join(cfg["work_dir"], f"events_n_duration_{dataset}_k{k}_{ilon}.{flon}.{ilat}.{flat}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh.csv"), index=False)
    
    # Example array with NaNs and integer values
    new_array = np.full((events_da.shape[0],events_da.shape[1]+1), np.nan)
    new_array[:,:-1] = events_da
    array=new_array.transpose().flatten()

    # Identify the indices where NaN values are located
    nan_indices = np.where(np.isnan(array))[0]

    # Split the array at these indices, ignoring the NaNs
    sub_arrays = np.split(array, nan_indices)
    # Filter out any sub-arrays that are empty, contain only NaNs, or have length 1
    filtered_sub_arrays = [sub_array[~np.isnan(sub_array)] for sub_array in sub_arrays if len(sub_array[~np.isnan(sub_array)]) > 1]

    logger.info(filtered_sub_arrays)
    states = list(np.arange(1,k+1))

    # Extract evaluation data
    # N days:
    ndays = int(yearly_intrusion_days.sum().values)
    # Trend
    trend_out=[slope, pval]
    out_dict = {"ndays": ndays, "trend": trend_out}
    if season != "ann" and "mjj" not in season:
        bxp = [list(i) for i in bxp]
        out_dict["bxp"] = bxp
    data_serializable = {key: value.tolist() if isinstance(value, np.ndarray) else value for key, value in out_dict.items()}
    print(data_serializable)
    if not os.path.isdir(os.path.join(cfg["work_dir"],"evaluation")):
        os.makedirs(os.path.join(cfg["work_dir"],"evaluation"))
    with open(os.path.join(cfg["work_dir"],"evaluation",f"{dataset}_{season}_{ilon}.{flon}.{ilat}.{flat}_{period.replace('-','').replace('.','-')}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}.json"), "w") as f:
        json.dump(data_serializable, f)

    # # Initialize the transition matrix
    # transition_matrix_total = np.zeros((len(states), len(states)))

    # for cleaned_array in filtered_sub_arrays:
    #     # Count transitions
    #     transition_matrix = np.zeros((len(states), len(states)))
    #     for (current_state, next_state) in zip(cleaned_array[:-1], cleaned_array[1:]):
    #         transition_matrix[int(current_state)-1, int(next_state)-1] += 1
    #     transition_matrix_total += transition_matrix
        
    
    # definitive_transition_matrix_total = np.full((transition_matrix_total.shape[0],transition_matrix_total.shape[0]+1),np.nan)
    # definitive_transition_matrix_total[:,:-1] = transition_matrix_total
    # def count_pattern_in_array(pattern, array):
    #     # Length of pattern and array
    #     pattern_len = len(pattern)
    #     array_len = len(array)
        
    #     # Initialize count of occurrences
    #     count = 0
        
    #     # Loop through the array, checking for the pattern
    #     for i in range(array_len - pattern_len + 1):
    #         # Compare subarray to pattern
    #         subarray = array[i:i + pattern_len]
    #         if all(np.isnan(a) and np.isnan(b) or a == b for a, b in zip(subarray, pattern)):
    #             count += 1
                
    #     return count
    
    # for it in np.arange(1, transition_matrix_total.shape[0]+1):
    #     definitive_transition_matrix_total[it-1,-1] = count_pattern_in_array([it, np.nan], array)
    # # Plot the transition matrix as a heatmap
    # plt.figure()
    # sns.heatmap(definitive_transition_matrix_total, annot=True, cmap="Blues", xticklabels=states+["event end"], yticklabels=states)
    # plt.xlabel("Next State")
    # plt.ylabel("Current State")
    # plt.title("Transition Probability Heatmap")
    # plt.savefig(os.path.join(cfg["plot_dir"], f"{dataset}_ind{indicators}_FigureS8{seasons_dict[season]}.png"))
    return events_da