import iris
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import numpy as np
import os
import xarray as xr
import logging
from datetime import datetime, timedelta
import cartopy
from matplotlib.gridspec import GridSpec
from esmvalcore.preprocessor import rolling_window_statistics, extract_region, extract_month, extract_time, extract_season, climate_statistics
from WCD_shared import plot_area_of_impact, regrid_longitude_coord, extract_months_xarray
from scipy.stats import t
import matplotlib.path as mpath


logger = logging.getLogger(__name__)
FORMAT = "%(asctime)-15s %(message)s"
logging.basicConfig(format=FORMAT)
logging.getLogger().setLevel(logging.INFO)


if 'BSC_MACHINE' in os.environ and os.environ['BSC_MACHINE'] == 'mn5':
    logger.info('Setting cartopy config for {} machine'.format(os.environ['BSC_MACHINE']))
    cartopy.config['data_dir'] = '/gpfs/projects/bsc32/bsc032259/data/cartopy/'

current_dir = os.path.dirname(os.path.abspath(__file__))


def plot_clustering(var_name, k, period, season, bounds, identifier, minimum_intrusion_points_pcnt, cluster_vars, region, sahara_clim_roll, sahara_region, ta_levels, gh_levels, level=None, precursor_day=-1, var_from="", identification_season=0, cluster_clim_only=True, significance=False):
    theta = np.linspace(0, 2*np.pi, 100)
    map_circle = mpath.Path(np.vstack([np.sin(theta), np.cos(theta)]).T * 0.5 + [0.5, 0.5])
    if identification_season == 0:
        identification_season = season
    ilon,flon,ilat,flat = bounds
    
    clusters_vector_file = f"{current_dir}/files/postpro_files/{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"
    intrusion_dates_file = os.path.join(current_dir,f"files/postpro_files/intrusion_dates/{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{identification_season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.txt")
    logger.info(clusters_vector_file)

    # SEND TO .JSON? 
    var_files = {"psl":f"{current_dir}/files/psl_r720x360_195001-202212_Sahara.nc",
                "theta": f"{current_dir}/files/avgtap925-700_1959-2022_01-12_fixed2.nc",
                "thick": f"{current_dir}/files/ghthick1000-500_1959-2022_01-12_fixed2.nc",
                "w": f"{current_dir}/files/w_1961-2022_01-12.nc",
                "ta": f"{current_dir}/files/ta_1981-2010_01-12.nc",
                "g": f"{current_dir}/files/gz_1959-2022_01-12.nc",
                "ua": f"{current_dir}/files/ua_1981-2010_01-12.nc",
                "aod": f"{current_dir}/files/od550aero_2001-2022.nc",
                "hus": f"{current_dir}/files/hus_1981-2010_01-12.nc",
                "g200": f"{current_dir}/files/g200_1959-2022_01-12.nc",
                "g300": f"{current_dir}/files/g300_1959-2022_01-12.nc",
                "g500": f"{current_dir}/files/g500_1959-2022_01-12.nc",
                "g150": f"{current_dir}/files/g150_1959-2022_01-12.nc",
                "va850": f"{current_dir}/files/merged/va850_1959-2022_01-12.nc",
                "ua850": f"{current_dir}/files/merged/ua850_1959-2022_01-12.nc",
                "va200": f"{current_dir}/files/va200_1959-2022_01-12.nc",
                "ua200": f"{current_dir}/files/ua200_1959-2022_01-12.nc",
                "va300": f"{current_dir}/files/va300_1959-2022_01-12.nc",
                "ua300": f"{current_dir}/files/ua300_1959-2022_01-12.nc",
                "tas": f"{current_dir}/files/tas_1959-2022_01-12_2.nc"}
    if gh_levels!="OG":
        var_files["thick"] = f"{current_dir}/files/merged/formatted/gh_{gh_levels}_1959-2022_01-12.nc"
    if ta_levels != "OG":
        var_files["thick"] = f"{current_dir}/files/merged/formatted/avgtap_{ta_levels}_1959-2022_01-12.nc"
    var_file = var_files[var_name]

    if season == "ann":
        desired_months = np.arange(1,13)
    elif isinstance(season, str):
        allmonths = "jfmamjjasond" * 2
        sstart = allmonths.index(season)
        desired_months = [i for i in range(sstart+1,sstart+1+len(season))]
        for idd,d in enumerate(desired_months):
            if d>12:
                desired_months[idd] -= 12
        desired_months.sort()
    elif isinstance(season, int) or isinstance(season, np.int64):
        desired_months = [season]

    # LOAD CLUSTERS, INTRUSION DATES, AND VAR TO BE CENTROIDED
    clusters_vector = xr.open_dataarray(clusters_vector_file)
    if clusters_vector.shape[0]==0:
        return 0 
    with open(intrusion_dates_file, "r") as f:
        intrusion_dates = [line.rstrip('\n') for line in f]
    var = iris.load_cube(var_file) # Centroided variable
    years = [i.year for i in var.coord("time").units.num2date(var.coord("time").points)] # Reference from VAR
    clusters_common_years = list(filter(lambda x: years[0]<=int(x.split("-")[0])<=years[-1], clusters_vector.time.values))
    clusters_vector = clusters_vector.sel(time=clusters_common_years)
    intrusion_dates = [i for i in intrusion_dates if years[0]<=int(i.split("-")[0])<=years[-1]]

    if precursor_day != -1:
        def convert_to_precursor(time_arr, intrusion_dates_og):
            out = [i for i in [(datetime(year=int(date[:4]), month=int(date[5:7]), day=int(date[8:]))-timedelta(days=precursor_day)).strftime('%Y-%m-%d') for date in time_arr] if precursor_day==0 or i not in intrusion_dates_og]
            return out

        def precursor_day_finder(intrusion_dates):
            # allmonths = "jfmamjjasond" * 2
            # sstart = allmonths.index(season)
            # for j in range(precursor_day+2):
            #     logger.info(j)
            #     intrusion_dates = [i for i in intrusion_dates if f"-0{sstart+1}-0{j}" not in i]
            diffs = []
            for date0,date1 in zip(["1900-01-01"]+intrusion_dates,intrusion_dates+["9999-12-31"]):
                diffs.append((datetime(year=int(date1[:4]), month=int(date1[5:7]), day=int(date1[8:]))-datetime(year=int(date0[:4]), month=int(date0[5:7]), day=int(date0[8:]))).days)
            diffs = np.array(diffs)[:-1]
            # Find the indices where the differences are not equal to 1
            start_indices = np.where(diffs != 1)[0] # AIXÒ NO ESTÂ BË
            
            # Spell duration
            spell_duration=np.array(list(start_indices[1:])+[len(intrusion_dates)])-start_indices
            # List of spell first day
            t0=[intrusion_dates[start_index] for start_index in start_indices]
            intrusion_dates_og = intrusion_dates.copy()
            t_minus = convert_to_precursor(t0, intrusion_dates_og)
            intrusion_dates = t_minus

            return intrusion_dates_og, intrusion_dates, t0

    if precursor_day != -1:
        intrusion_dates_og, intrusion_dates, t0 = precursor_day_finder(intrusion_dates)
        t0_season = [i for i in t0 if int(i.split("-")[1]) in desired_months]
        clusters_vector = clusters_vector.sel(time=t0_season)

    if level == None:
        var_clim=xr.open_dataarray(f"{current_dir}/files/postpro_files/composites/composite-{var_name}-mean_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
    else:
        var_clim=xr.open_dataarray(f"{current_dir}/files/postpro_files/composites/composite-{var_name}{level}-mean_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
    format = {"units": {r"$\overline{\theta}_{700-925}$": "K",r"$\Delta GH_{500-1000}$": "m", "theta": "K", "thick": "m", "psl": "Pa", "g": "m", "w": "m/s", "ua": "m/s", "ta": "K", "tas": "K", "aod": "AOD", "hus": "kg/kg", "g200": "m2 s-2", "g300": "m", "g500": "m", "g150": "m", "va850": "m2 s-2", "ua850": "m2 s-2", "va200": "m/s", "ua200": "m/s", "va300": "m/s", "ua300": "m/s"}} 

    if level ==  None:
        level=""

    if var_name == "theta":
        if season == "jja":
            vmax = 6
        elif season == "djf":
            vmax=8
        elif season == "son":
            vmax=8
        elif season == "mam":
            vmax = 10
        else:
            vmax=8
    elif var_name == "thick":
        if season == "jja":
            vmax = 80
        elif season == "djf":
            vmax=150
        elif season == "son": 
            vmax=150
        elif season == "mam":
            vmax=200
        else:
            vmax=150
    elif var_name == "ta":
        vmax=8
    else:
        vmax=100

    if not os.path.isdir(f"{current_dir}/plots/composites/"):
        os.makedirs(f"{current_dir}/plots/composites/")

    # Plot the centroids of the VAR. This takes time, should there be intermediary files?
    fig = plt.figure(figsize=(5,2.7*len(np.unique(clusters_vector))), dpi=200)
    gs = GridSpec(len(np.unique(clusters_vector)), 1, bottom=0.09, top=0.96)# hspace=0.0001)
    
    for cluster_n in np.unique(clusters_vector):
        dates_og = clusters_vector.where(clusters_vector==cluster_n, drop=True).time
        if precursor_day != -1:
            dates = convert_to_precursor(dates_og.values, intrusion_dates_og)
        else:
            dates = dates_og.copy()
        if len(dates) == 0:
            ax = plt.subplot(gs[cluster_n, :], projection=ccrs.PlateCarree())
            ax.set_title(f"All precursor days {precursor_day} from IT{cluster_n+1}\ncoincide with other intrusion days")
            continue
        if level == None:
            var_cluster_mean=xr.open_dataarray(f"{current_dir}/files/postpro_files/composites/composite-{var_name}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
        else:
            var_cluster_mean=xr.open_dataarray(f"{current_dir}/files/postpro_files/composites/composite-{var_name}{level}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
        if var_cluster_mean.lon.values[-1]-var_cluster_mean.lon.values[0] > 350:
            ax = plt.subplot(gs[cluster_n, :], projection=ccrs.PlateCarree())
            # ax = plt.subplot(gs[cluster_n, :], projection=ccrs.NorthPolarStereo())
            # ax.set_extent([-180, 180, 15, 90], crs=ccrs.PlateCarree())
            # ax.set_boundary(map_circle, transform=ax.transAxes)
        else:
            ax = plt.subplot(gs[cluster_n, :], projection=ccrs.PlateCarree())
            ax.set_extent([-30, 40, 15, 50], crs=ccrs.PlateCarree())
            var_cluster_mean=xr.DataArray.from_iris(extract_region(var_cluster_mean.to_iris(), -30, 40, 15, 50))
        cax = var_cluster_mean.plot(cmap=plt.cm.RdBu_r, ax=ax, vmin=-1*vmax, vmax=vmax, transform=ccrs.PlateCarree(), add_colorbar=False)
        z_min = np.ceil(var_clim.min() * 2) / 2
        z_max = np.floor(var_clim.max() * 2) / 2
        step = (z_max - z_min) / 9
        step = np.round(step * 2) / 2 
        levels = np.arange(z_min, z_max + step, step)[:10]
        ct = var_clim.plot.contour(levels=levels, colors='black', linewidths=1, transform=ccrs.PlateCarree())
        contour_levels = ct.levels[::3]
        ax.clabel(ct, levels=contour_levels, inline=True, fontsize=10)
        area_of_impact = xr.open_dataarray(os.path.join(current_dir,f"files/postpro_files/area-of-impact-n{cluster_n+1}_clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"))
        ax = plot_area_of_impact(ax, area_of_impact, "dashed")
        ax.coastlines()
        if var_name == "theta":
            display_var_name = r"$\overline{\theta}_{700-925}$"
        elif var_name == "thick":
            display_var_name = r"$\Delta GH_{500-1000}$"
        else:
            display_var_name = var_name.upper()
        if precursor_day != -1:
            ax.set_title(f"{season.upper()} IT{cluster_n+1} {display_var_name}{level} DAY {-1*precursor_day} composite mean & anom.")
        else:
            ax.set_title(f"{season.upper()} IT{cluster_n+1} {display_var_name}{level} composite mean & anom.")
    cbar_ax = fig.add_axes([0.2, 0.06, 0.6, 0.01*5/len(np.unique(clusters_vector))])
    fig.colorbar(cax, label=format["units"][var_name], cax=cbar_ax, orientation="horizontal", extend="both")
    plt.tight_layout()
    if level == None:
        plt.savefig(f"{current_dir}/plots/composites/{var_from}from_{var_name}-anom-clusters{k}_{season}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.png")#, bbox_inches="tight")
    else:
        plt.savefig(f"{current_dir}/plots/composites/{var_from}from_{var_name}{level}-anom-clusters{k}_{season}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.png", bbox_inches="tight")
    plt.close() 


def plot_quiver_clustering(var_name, quiver_level, k, period, season, bounds, identifier, minimum_intrusion_points_pcnt, cluster_vars, region, sahara_clim_roll, sahara_region, ta_levels, gh_levels, precursor_day=-1, var_from="", identification_season=0, cluster_clim_only=True, significance=False):
    if identification_season == 0:
        identification_season = season
    ilon,flon,ilat,flat = bounds
    # És el propi de l'estació
        
    clusters_vector_file = f"{current_dir}/files/postpro_files/{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"
    var_clim = xr.DataArray.from_iris(regrid_longitude_coord(extract_region(iris.load_cube(f"{current_dir}/files/postpro_files/composites/composite-{var_name}-mean_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"), -40, 50, 15,70)))
    # És l'annual!
    intrusion_dates_file = os.path.join(current_dir,f"files/intrusion_dates/{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{identification_season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh.txt")
    # intrusion_dates_file = os.path.join(current_dir,f"files/intrusion_dates/{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{identification_season}_min-intr-points-{minimum_intrusion_points_pcnt}.txt")
     
    clusters_vector = xr.open_dataarray(clusters_vector_file)
    
    format = {"var_names": {"g300": "GH 300 hPa","g850": "GH 850 hPa", "psl": "SLP"},"units": {r"$\overline{\theta}_{700-925}$": "K",r"$\Delta GH_{500-1000}$": "m", "theta": "K", "thick": "m", "psl": "Pa", "g": "m", "w": "m/s", "ua": "m/s", "ta": "K", "tas": "K", "aod": "AOD", "hus": "kg/kg", "g200": "m2 s-2", "g300": "m", "g500": "m",  "g850": "m", "g150": "m","va850": "m2 s-2" , "va200": "m/s", "ua200": "m/s"}} 
    
    if var_name == "g300":
        if season == "jja":
            vmax = 150
        elif season == "djf":
            vmax=250
        elif season == "son" or season == "mam":
            vmax=180
        else:
            vmax=200
    elif var_name == "ua200":
        if season == "jja":
            vmax = 15
        elif season == "djf":
            vmax=30
        elif season == "son" or season == "mam":
            vmax=25
        else:
            vmax=20
    elif var_name == "psl":
        if season == "jja":
            vmax = 900 
        elif season == "djf":
            vmax=1900
        elif season == "son" or season == "mam":
            vmax=1500
        else:
            vmax=1000
    elif var_name == "g850":
        if season == "jja":
            vmax = 80
        elif season == "son" or season == "djf":
            vmax=120
        elif season == "mam":
            vmax=100
        else:
            vmax=100
    else:
        vmax=100

    if int(quiver_level) == 850:
        scale_factors={"djf": 1.5, "jja": 0.8, "mam": 1.3, "son": 1.3, "mjj": 1.1, "amjja": 1.1}
        density = {"djf": 3, "jja": 3, "mam": 3, "son": 3, "mjj": 3, "amjja":3} 
        if season != "jja":
            ws_value = 10
        else:
            ws_value = 5
    if int(quiver_level) == 300:
        scale_factors={"djf": 4, "jja": 3, "mam": 4, "son": 4, "mjj": 4, "amjja":4}
        density = {"djf": 5, "jja": 4, "mam": 5, "son": 5, "mjj": 5, "amjja": 5} 
        if season != "jja":
            ws_value = 20
        else:
            ws_value = 10
    # PLot the centroids of the VAR. This takes time, should there be intermediary files?
    fig = plt.figure(figsize=(5,2.7*len(np.unique(clusters_vector))), dpi=200)
    gs = GridSpec(len(np.unique(clusters_vector)), 1, bottom=0.09, top=0.96)# hspace=0.0001)
    
    for cluster_n in np.unique(clusters_vector):
        var_cluster_mean = xr.DataArray.from_iris(regrid_longitude_coord(extract_region(iris.load_cube(f"{current_dir}/files/postpro_files/composites/composite-{var_name}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"), -40, 50, 15, 70)))
        var_cluster_mean_u = xr.DataArray.from_iris(regrid_longitude_coord(extract_region(iris.load_cube(f"{current_dir}/files/postpro_files/composites/nomask-composite-ua{quiver_level}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"), -40, 50, 15, 70)))
        masked_u = xr.DataArray.from_iris(regrid_longitude_coord(extract_region(iris.load_cube(f"{current_dir}/files/postpro_files/composites/composite-ua{quiver_level}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"), -40, 50, 15, 70)))
        var_cluster_mean_v = xr.DataArray.from_iris(regrid_longitude_coord(extract_region(iris.load_cube(f"{current_dir}/files/postpro_files/composites/nomask-composite-va{quiver_level}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"), -40, 50, 15, 70)))
        masked_v = xr.DataArray.from_iris(regrid_longitude_coord(extract_region(iris.load_cube(f"{current_dir}/files/postpro_files/composites/composite-va{quiver_level}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"), -40, 50, 15, 70)))
        
        nan_mask = np.isnan(masked_u) & np.isnan(masked_v)
        var_cluster_mean_v = xr.where(nan_mask,np.nan,var_cluster_mean_v)
        var_cluster_mean_u = xr.where(nan_mask,np.nan,var_cluster_mean_u)

        lon = var_cluster_mean_u.lon
        lat = var_cluster_mean_u.lat

        # if var_cluster_mean.lon.values[-1]-var_cluster_mean.lon.values[0] > 350:
        #     ax = plt.subplot(gs[cluster_n, :], projection=ccrs.NorthPolarStereo())
        #     ax.set_boundary(map_circle, transform=ax.transAxes)
        # else:
        ax = plt.subplot(gs[cluster_n, :], projection=ccrs.PlateCarree())
        var_cluster_mean=xr.DataArray.from_iris(extract_region(var_cluster_mean.to_iris(), -40, 50, 15, 70))
        contourf = var_cluster_mean.plot(cmap=plt.cm.RdBu_r, ax=ax, vmin=-1*vmax, vmax=vmax, transform=ccrs.PlateCarree(), add_colorbar=False)
        # contourf = var_cluster_mean.plot(cmap=plt.cm.RdBu_r, ax=ax, cbar_kwargs={"label": format["units"][var_name]}, vmin=-1*vmax, vmax=vmax, transform=ccrs.PlateCarree(), add_colorbar=False)
        z_min = np.ceil(var_clim.min() * 2) / 2
        z_max = np.floor(var_clim.max() * 2) / 2
        step = (z_max - z_min) / 9
        step = np.round(step * 2) / 2 
        levels = np.arange(z_min, z_max + step, step)[:10]
        ct = var_clim.plot.contour(levels=levels, colors='gray', linewidths=1, transform=ccrs.PlateCarree())
        
        # ct = var_clim.plot.contour(levels=np.linspace(var_clim.min(), var_clim.max(), 10), colors='black', linewidths=1, transform=ccrs.PlateCarree())
        contour_levels = ct.levels[::3]
        ax.clabel(ct, levels=contour_levels, inline=True, fontsize=8)
        scale_factor = scale_factors[season] # smaller makes arrow larger
        q = ax.quiver(lon[::density[season]], lat[::density[season]], var_cluster_mean_u[::density[season], ::density[season]], var_cluster_mean_v[::density[season], ::density[season]], scale=scale_factor, color='black', scale_units='xy') #, width=0.0012, headwidth=5)
        # Key  
        sf = ws_value 
        label = fr'${sf}  m/s$'
        area_of_impact = xr.open_dataarray(os.path.join(current_dir,f"files/postpro_files/area-of-impact-n{cluster_n+1}_clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"))
        ax = plot_area_of_impact(ax, area_of_impact, "dashed")
        ax.set_extent([-40, 50, 15, 70], crs=ccrs.PlateCarree())
        if cluster_n == 0:
            # qk = ax.quiverkey(q, 0.1, 0.97, sf, label, labelpos='N', coordinates='figure')
            qk = ax.quiverkey(q, 0.035, 1.03, sf, label, labelpos='N', coordinates='axes')
            ax.set_title(f"        {season.upper()} IT{cluster_n+1} {format['var_names'][var_name]} & wind {quiver_level} hPa")
        else:
            ax.set_title(f"{season.upper()} IT{cluster_n+1} {format['var_names'][var_name]} & wind {quiver_level} hPa")
        ax.coastlines()
        # Set the boundary to be circular by applying this mask
        # ax.set_boundary(circle, transform=ax.transAxes)
        if var_name == "theta":
            display_var_name = r"$\overline{\theta}_{700-925}$"
        elif var_name == "thick":
            display_var_name = r"$\Delta GH_{500-1000}$"
        else:
            display_var_name = var_name.upper()
            #     ax.set_title(f"IT{cluster_n+1} {display_var_name}{level} composite std_dev")
    cbar_ax = fig.add_axes([0.2, 0.06, 0.6, 0.01*5/len(np.unique(clusters_vector))])

    fig.colorbar(contourf, label=format["units"][var_name], cax=cbar_ax, orientation="horizontal", extend="both")
    plt.tight_layout()
    plt.savefig(f"{current_dir}/plots/composites/paper_plot_{season}_{var_name}_quiver-{quiver_level}.png", bbox_inches="tight")
    plt.close()

def compute_composites(var_name, k, period, season, bounds, identifier, minimum_intrusion_points_pcnt, cluster_vars, region, sahara_clim_roll, sahara_region, ta_levels, gh_levels, level=None, precursor_day=-1, var_from="", identification_season=0, cluster_clim_only=True, significance=False):
    theta = np.linspace(0, 2*np.pi, 100)
    map_circle = mpath.Path(np.vstack([np.sin(theta), np.cos(theta)]).T * 0.5 + [0.5, 0.5])
    if identification_season == 0:
        identification_season = season
    ilon,flon,ilat,flat = bounds
    clusters_vector_file = f"{current_dir}/files/postpro_files/{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc"
    intrusion_dates_file = os.path.join(current_dir,f"files/postpro_files/intrusion_dates/{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{identification_season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.txt")
    logger.info(clusters_vector_file)
     
    var_files = {"psl":f"{current_dir}/files/psl_r720x360_195001-202212_Sahara.nc",
                "theta": f"{current_dir}/files/avgtap925-700_1959-2022_01-12_fixed2.nc",
                "thick": f"{current_dir}/files/ghthick1000-500_1959-2022_01-12_fixed2.nc",
                "w": f"{current_dir}/files/w_1961-2022_01-12.nc",
                "ta": f"{current_dir}/files/ta_1981-2010_01-12.nc",
                "g": f"{current_dir}/files/gz_1959-2022_01-12.nc",
                "ua": f"{current_dir}/files/ua_1981-2010_01-12.nc",
                "aod": f"{current_dir}/files/od550aero_2001-2022.nc",
                "hus": f"{current_dir}/files/hus_1981-2010_01-12.nc",
                "g200": f"{current_dir}/files/g200_1959-2022_01-12.nc",
                "g300": f"{current_dir}/files/g300_1959-2022_01-12.nc",
                "g500": f"{current_dir}/files/g500_1959-2022_01-12.nc",
                "g150": f"{current_dir}/files/g150_1959-2022_01-12.nc",
                "va850": f"{current_dir}/files/merged/va850_1959-2022_01-12.nc",
                "ua850": f"{current_dir}/files/merged/ua850_1959-2022_01-12.nc",
                "va200": f"{current_dir}/files/va200_1959-2022_01-12.nc",
                "ua200": f"{current_dir}/files/ua200_1959-2022_01-12.nc",
                "va300": f"{current_dir}/files/va300_1959-2022_01-12.nc",
                "ua300": f"{current_dir}/files/ua300_1959-2022_01-12.nc",
                "tas": f"{current_dir}/files/tas_1959-2022_01-12_2.nc"}
    if gh_levels!="OG":
        var_files["thick"] = f"{current_dir}/files/merged/formatted/gh_{gh_levels}_1959-2022_01-12.nc"
    if ta_levels != "OG":
        var_files["thick"] = f"{current_dir}/files/merged/formatted/avgtap_{ta_levels}_1959-2022_01-12.nc"
    var_file = var_files[var_name]

    if season == "ann":
        desired_months = np.arange(1,13)
    elif isinstance(season, str):
        allmonths = "jfmamjjasond" * 2
        sstart = allmonths.index(season)
        desired_months = [i for i in range(sstart+1,sstart+1+len(season))]
        for idd,d in enumerate(desired_months):
            if d>12:
                desired_months[idd] -= 12
        desired_months.sort()
    elif isinstance(season, int) or isinstance(season, np.int64):
        desired_months = [season]

    clusters_vector = xr.open_dataarray(clusters_vector_file)
    if clusters_vector.shape[0]==0:
        return 0 
    with open(intrusion_dates_file, "r") as f:
        intrusion_dates = [line.rstrip('\n') for line in f]
    var = iris.load_cube(var_file) # Centroided variable
    years = [i.year for i in var.coord("time").units.num2date(var.coord("time").points)] # Reference from VAR
    clusters_common_years = list(filter(lambda x: years[0]<=int(x.split("-")[0])<=years[-1], clusters_vector.time.values))
    clusters_vector = clusters_vector.sel(time=clusters_common_years)
    intrusion_dates = [i for i in intrusion_dates if years[0]<=int(i.split("-")[0])<=years[-1]]

    if precursor_day != -1:
        def convert_to_precursor(time_arr, intrusion_dates_og):
            out = [i for i in [(datetime(year=int(date[:4]), month=int(date[5:7]), day=int(date[8:]))-timedelta(days=precursor_day)).strftime('%Y-%m-%d') for date in time_arr] if precursor_day==0 or i not in intrusion_dates_og]
            return out

        def precursor_day_finder(intrusion_dates):
            # allmonths = "jfmamjjasond" * 2
            # sstart = allmonths.index(season)
            # for j in range(precursor_day+2):
            #     logger.info(j)
            #     intrusion_dates = [i for i in intrusion_dates if f"-0{sstart+1}-0{j}" not in i]
            diffs = []
            for date0,date1 in zip(["1900-01-01"]+intrusion_dates,intrusion_dates+["9999-12-31"]):
                diffs.append((datetime(year=int(date1[:4]), month=int(date1[5:7]), day=int(date1[8:]))-datetime(year=int(date0[:4]), month=int(date0[5:7]), day=int(date0[8:]))).days)
            diffs = np.array(diffs)[:-1]
            # diffs = np.diff(dates_int)
            # Find the indices where the differences are not equal to 1
            start_indices = np.where(diffs != 1)[0] # AIXÒ NO ESTÂ BË
            
            # Spell duration
            spell_duration=np.array(list(start_indices[1:])+[len(intrusion_dates)])-start_indices
            # List of spell first day
            t0=[intrusion_dates[start_index] for start_index in start_indices]
        
            intrusion_dates_og = intrusion_dates.copy()
            
            t_minus = convert_to_precursor(t0, intrusion_dates_og)
            
            intrusion_dates = t_minus

            return intrusion_dates_og, intrusion_dates, t0

    if region != None:
        var = xr.DataArray.from_iris(regrid_longitude_coord(extract_region(var, region[0],region[1],region[2],region[3])))
    else:
        var = xr.DataArray.from_iris(regrid_longitude_coord(var))

    if "levels" in var.coords:
        if level == None:
            raise ValueError("you need to specify a level")
        else:
            var = var.sel(levels=level)
            # cl_t0=find_t0(list(dates_og.values))

    og_times = var.time.values
    all_time_vector = [str(i).split(" ")[0] for i in og_times]
    var_dumb_time = var.assign_coords(time=all_time_vector)
    # This keeps all the days but masks those with intrusion
    mask = var_dumb_time['time'].isin(intrusion_dates).broadcast_like(var_dumb_time).assign_coords(time=og_times).to_iris()
    var_cube = var.assign_coords(time=og_times).to_iris()
    var_no_intrusion_days = iris.util.mask_cube(var_cube, mask)

    # Compute a FULL anomaly cube.  This takes time, substitute by an intermediary file?
    var = var.assign_coords(time=all_time_vector)
    windowed_clim = climate_statistics(rolling_window_statistics(var_no_intrusion_days, "time", "mean", 7), "mean", "daily")
    var = var.assign_coords(time=og_times)
    # STACK AND ANOMALY    
    stacked_clim = []
    dts=var.to_iris().coord("time").units.num2date(var.to_iris().coord("time").points)
    for year in np.arange(int(dts[0].year), int(dts[-1].year)+1):
        clim_data = list(windowed_clim.data)
        if extract_time(var.to_iris(), year, 1, 1, year+1, 1, 1).shape[0] != 366:
            del clim_data[59]
        stacked_clim += clim_data
    clim_cube = var.to_iris().copy(np.array(stacked_clim))
    clim_cube = xr.DataArray.from_iris(clim_cube)

    if not os.path.isdir(f"{current_dir}/files/postpro_files/composites"):
        os.makedirs(f"{current_dir}/files/postpro_files/composites")
    logger.info(f"SAVING {current_dir}/files/postpro_files/composites/clim-composite-{var_name}-mean_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
    if level == None:
        clim_cube.to_netcdf(f"{current_dir}/files/postpro_files/composites/clim-composite-{var_name}-mean_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
    else:
        clim_cube.to_netcdf(f"{current_dir}/files/postpro_files/composites/clim-composite-{var_name}{level}-mean_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")

    anom_var = var-clim_cube

    # renew instrusion_dates
    if precursor_day == -1:
        anom_var = anom_var.assign_coords(time=[str(i).split(" ")[0] for i in anom_var.time.values]).sel(time=intrusion_dates)
        anom_var = extract_months_xarray(anom_var, desired_months)
    else:
        intrusion_dates_og, intrusion_dates, t0 = precursor_day_finder(intrusion_dates)
        t0_season = [i for i in t0 if int(i.split("-")[1]) in desired_months]
        intrusion_dates = convert_to_precursor(t0_season, intrusion_dates_og)
        anom_var = anom_var.assign_coords(time=[str(i).split(" ")[0] for i in anom_var.time.values]).sel(time=intrusion_dates)
        clusters_vector = clusters_vector.sel(time=t0_season)

    if season != "ann" and isinstance(season, str):
        logger.info(f"extracting season {season}")
        var = xr.DataArray.from_iris(extract_season(var.assign_coords(time=og_times).to_iris(), season))

    elif isinstance(season, int):
        logger.info(f"extracting month {season}")
        var = xr.DataArray.from_iris(extract_month(var.assign_coords(time=og_times).to_iris(), season))
    else:
        pass
    var_clim=var.mean(axis=0)   
    logger.info(f"SAVING {current_dir}/files/postpro_files/composites/composite-{var_name}-mean_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
    if level == None:
        var_clim.to_netcdf(f"{current_dir}/files/postpro_files/composites/composite-{var_name}-mean_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
    else:
        var_clim.to_netcdf(f"{current_dir}/files/postpro_files/composites/composite-{var_name}{level}-mean_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
    format = {"units": {r"$\overline{\theta}_{700-925}$": "K",r"$\Delta GH_{500-1000}$": "m", "theta": "K", "thick": "m", "psl": "Pa", "g": "m", "w": "m/s", "ua": "m/s", "ta": "K", "tas": "K", "aod": "AOD", "hus": "kg/kg", "g200": "m2 s-2", "g300": "m", "g500": "m", "g150": "m", "va850": "m2 s-2", "ua850": "m2 s-2", "va200": "m/s", "ua200": "m/s", "va300": "m/s", "ua300": "m/s"}} 

    def compute_pvalue(n_var_cluster, var_cluster_mean_t, var_std_t, var_cluster_std_t):
        n_var = (years[-1]-years[0]+1)*7
        t_stat = (var_cluster_mean_t) / (var_cluster_std_t/np.sqrt(n_var_cluster))
        df = n_var_cluster-1
        p_value = 2 * t.sf(np.abs(t_stat), df)
        return p_value, df

    if level ==  None:
        level=""

    for cluster_n in np.unique(clusters_vector):
        dates_og = clusters_vector.where(clusters_vector==cluster_n, drop=True).time
        if precursor_day != -1:
            dates = convert_to_precursor(dates_og.values, intrusion_dates_og)
        else:
            dates = dates_og.copy()

        var_anom_std = 0 #clim_cube.assign_coords(time=[str(i).split(" ")[0] for i in clim_cube.time.values]).sel(time=dates).std(axis=0)
        var_cluster_comp = anom_var.sel(time=dates)
        var_cluster_mean = (var_cluster_comp).mean(axis=0)
        var_cluster_std = (var_cluster_comp).std(axis=0)
        n_var_cluster = var_cluster_comp.count(dim="time")
        p_values, df = compute_pvalue(n_var_cluster, var_cluster_mean, var_anom_std, var_cluster_std)
        if "ua" in var_name or "va" in var_name:
            var_cluster_mean_nomask=var_cluster_mean.copy()
        if significance:
            var_cluster_mean = xr.where(p_values<0.05, var_cluster_mean, np.nan)
            if level == None:
                var_cluster_mean.to_netcdf(f"{current_dir}/files/postpro_files/composites/composite-{var_name}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
                if "ua" in var_name or "va" in var_name:
                    var_cluster_mean_nomask.to_netcdf(f"{current_dir}/files/postpro_files/composites/nomask-composite-{var_name}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
            else:
                var_cluster_mean.to_netcdf(f"{current_dir}/files/postpro_files/composites/composite-{var_name}{level}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
                if "ua" in var_name or "va" in var_name:
                    var_cluster_mean_nomask.to_netcdf(f"{current_dir}/files/postpro_files/composites/nomask-composite-{var_name}{level}-anom_IT{cluster_n+1}_{var_from}clusters-k{k}_{'.'.join(cluster_vars)}_{ilon}.{flon}.{ilat}.{flat}_intrusions-clim-dates{identifier}_{period.replace('-','').replace('.','-')}_{season}_min-intr-points-{minimum_intrusion_points_pcnt}_sahara-{sahara_region[0]}.{sahara_region[1]}.{sahara_region[2]}.{sahara_region[3]}-roll.{sahara_clim_roll}-thresh_ta{ta_levels}_gh{gh_levels}.nc")
